package paxos

//
// Paxos library, to be included in an application.
// Multiple applications will run, each including
// a Paxos peer.
//
// Manages a sequence of agreed-on values.
// The set of peers is fixed.
// Copes with network failures (partition, msg loss, &c).
// Does not store anything persistently, so cannot handle crash+restart.
//
// The application interface:
//
// px = paxos.Make(peers []string, me string)
// px.Start(seq int, v interface{}) -- start agreement on new instance
// px.Status(seq int) (decided bool, v interface{}) -- get info about an instance
// px.Done(seq int) -- ok to forget all instances <= seq
// px.Max() int -- highest instance seq known, or -1
// px.Min() int -- instances before this seq have been forgotten
//

import "net"
import "net/rpc"
import "log"
import "os"
import "syscall"
import "sync"
import "fmt"
import "math/rand"
import "time"


type N struct {
  T time.Time
  Server string
}

type InstanceState struct {
  Np N
  Na N
  Va interface{}
  Decided bool
  DecidedValue interface{}
}

type Paxos struct {
  mu sync.Mutex
  l net.Listener
  dead bool
  unreliable bool
  rpcCount int
  peers []string
  me int // index into peers[]
  instances map[int]*InstanceState
  lastDeletedInstance int
  highestSeenInstance int
  latestDone []int
}

type PrepareArgs struct {
  Seq int
  Pn N
}

type PrepareReply struct {
  Na N
  Va interface{}
  Ok bool
  LatestDone int
}

type AcceptArgs struct {
  Seq int
  An N
  Av interface{}
}

type AcceptReply struct {
  Ok bool
  LatestDone int
}

type DecidedArgs struct {
  Seq int
  DecidedValue interface{}
}

type DecidedReply struct {}

//
// call() sends an RPC to the rpcname handler on server srv
// with arguments args, waits for the reply, and leaves the
// reply in reply. the reply argument should be a pointer
// to a reply structure.
//
// the return value is true if the server responded, and false
// if call() was not able to contact the server. in particular,
// the replys contents are only valid if call() returned true.
//
// you should assume that call() will time out and return an
// error after a while if it does not get a reply from the server.
//
// please use call() to send all RPCs, in client.go and server.go.
// please do not change this function.
//
func call(srv string, name string, args interface{}, reply interface{}) bool {
  c, err := rpc.Dial("unix", srv)
  if err != nil {
    err1 := err.(*net.OpError)
    if err1.Err != syscall.ENOENT && err1.Err != syscall.ECONNREFUSED {
      fmt.Printf("paxos Dial() failed: %v\n", err1)
    }
    return false
  }
  defer c.Close()
    
  err = c.Call(name, args, reply)
  if err == nil {
    return true
  }

  fmt.Println(err)
  return false
}

func (px *Paxos) Prepare(args *PrepareArgs, reply *PrepareReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  px.addMapEntryIfNeeded(args.Seq)
  if greater(args.Pn, px.instances[args.Seq].Np) {
    px.instances[args.Seq].Np = args.Pn
    *reply = PrepareReply{px.instances[args.Seq].Na, px.instances[args.Seq].Va, true,
    px.latestDone[px.me]}
  } else {
    *reply = PrepareReply{N{}, nil, false, px.latestDone[px.me]}
  }
  return nil
}

func (px *Paxos) Decided(args *DecidedArgs, reply *DecidedReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  px.addMapEntryIfNeeded(args.Seq)
  px.instances[args.Seq].Decided = true
  px.instances[args.Seq].DecidedValue = args.DecidedValue
  return nil
}

func (px *Paxos) Accept(args *AcceptArgs, reply *AcceptReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  px.addMapEntryIfNeeded(args.Seq)
  if !greater(px.instances[args.Seq].Np, args.An) {
    px.instances[args.Seq].Np = args.An
    px.instances[args.Seq].Na = args.An
    px.instances[args.Seq].Va = args.Av
    *reply = AcceptReply{true, px.latestDone[px.me]}
  } else {
    *reply = AcceptReply{false, px.latestDone[px.me]}
  }
  return nil
}

func greater(one N, two N) bool {
  if one.T.After(two.T) {
    return true
  } else if one.T.Equal(two.T) {
    return one.Server > two.Server
  } else {
    return false
  }
}

func (px *Paxos) isMajority(num int) bool {
  return 2 * num > len(px.peers)
}

func (px *Paxos) addMapEntryIfNeeded(seq int) {
  if _, ok := px.instances[seq]; ok {
    return // already there
  }
  if seq > px.highestSeenInstance {
    px.highestSeenInstance = seq
  }
  px.instances[seq] = &InstanceState{}
}

func (px *Paxos) garbageCollect() {
  px.mu.Lock()
  defer px.mu.Unlock()
  minDone := px.latestDone[0]
  for i := 1; i < len(px.latestDone); i++ {
    if px.latestDone[i] < minDone {
      minDone = px.latestDone[i]
    }
  }
  for i := px.lastDeletedInstance + 1; i <= minDone; i++ {
    delete(px.instances, i)
  }
  px.lastDeletedInstance = minDone
}

func (px *Paxos) isDecided(seq int) bool {
  decided, _ := px.Status(seq)
  return decided
}

//
// the application wants paxos to start agreement on
// instance seq, with proposed value v.
// Start() returns right away; the application will
// call Status() to find out if/when agreement
// is reached.
//
func (px *Paxos) Start(seq int, value interface{}) {
  px.mu.Lock()
  px.addMapEntryIfNeeded(seq)
  px.mu.Unlock()
  go func() {
    for !px.isDecided(seq) && !px.dead {
      // fmt.Printf("Proposer Loop %d...\n", px.me)
      n := N{time.Now(), px.peers[px.me]}
      var largestPrepareReply PrepareReply
      isResponseNa := false
      numPrepareOk := 0
      prepareArgs := &PrepareArgs{seq, n}
      for i, peer := range px.peers {
        // fmt.Printf("Prepare to Peer %d...\n", i)
        prepareReply := &PrepareReply{}
        ok := true
        if i == px.me {
          px.Prepare(prepareArgs, prepareReply)
        } else {
          ok = call(peer, "Paxos.Prepare", prepareArgs, prepareReply)
          if ok {
            px.mu.Lock()
            px.latestDone[i] = prepareReply.LatestDone
            px.mu.Unlock()
          }
        }
        if ok && prepareReply.Ok {
          numPrepareOk += 1
          if greater(prepareReply.Na, largestPrepareReply.Na) {
            isResponseNa = true
            largestPrepareReply = *prepareReply
          }
        }
      }
      // fmt.Println("Prepare Majority Check...")
      // fmt.Printf("Prepare Ok: %d\n", numPrepareOk)

      if px.isMajority(numPrepareOk) {
        v := value
        if isResponseNa {
          v = largestPrepareReply.Va
        }
        numAcceptOk := 0
        acceptArgs := &AcceptArgs{seq, n, v}
        for i, peer := range px.peers {
          acceptReply := &AcceptReply{}
          ok := true
          if i == px.me {
            px.Accept(acceptArgs, acceptReply)
          } else {
            ok = call(peer, "Paxos.Accept", acceptArgs, acceptReply)
            if ok {
              px.mu.Lock()
              px.latestDone[i] = acceptReply.LatestDone
              px.mu.Unlock()
            }
          }
          if ok && acceptReply.Ok {
            numAcceptOk += 1
          }
        }
        // fmt.Println("Accept Majority Check...")
        // fmt.Printf("Accept Ok: %d\n", numAcceptOk)
        if px.isMajority(numAcceptOk) {
          decidedArgs := &DecidedArgs{seq, v}
          for i, peer := range px.peers {
            if i == px.me {
              px.Decided(decidedArgs, &DecidedReply{})
            } else {
              call(peer, "Paxos.Decided", decidedArgs, &DecidedReply{})
            }
          }
        }
      }
      // fmt.Println("Garbage Collection...")
      px.garbageCollect()
      time.Sleep(10 * time.Millisecond)
    }
    // fmt.Println("Proposer Done...")
  }()
}

//
// the application on this machine is done with
// all instances <= seq.
//
// see the comments for Min() for more explanation.
//
func (px *Paxos) Done(seq int) {
  px.mu.Lock()
  defer px.mu.Unlock()
  if seq > px.latestDone[px.me] {
    px.latestDone[px.me] = seq
  }
}

//
// the application wants to know the
// highest instance sequence known to
// this peer.
//
func (px *Paxos) Max() int {
  px.mu.Lock()
  defer px.mu.Unlock()
  return px.highestSeenInstance
}

//
// Min() should return one more than the minimum among z_i,
// where z_i is the highest number ever passed
// to Done() on peer i. A peers z_i is -1 if it has
// never called Done().
//
// Paxos is required to have forgotten all information
// about any instances it knows that are < Min().
// The point is to free up memory in long-running
// Paxos-based servers.
//
// Paxos peers need to exchange their highest Done()
// arguments in order to implement Min(). These
// exchanges can be piggybacked on ordinary Paxos
// agreement protocol messages, so it is OK if one
// peers Min does not reflect another Peers Done()
// until after the next instance is agreed to.
//
// The fact that Min() is defined as a minimum over
// *all* Paxos peers means that Min() cannot increase until
// all peers have been heard from. So if a peer is dead
// or unreachable, other peers Min()s will not increase
// even if all reachable peers call Done. The reason for
// this is that when the unreachable peer comes back to
// life, it will need to catch up on instances that it
// missed -- the other peers therefor cannot forget these
// instances.
// 
func (px *Paxos) Min() int {
  px.mu.Lock()
  defer px.mu.Unlock()
  return px.lastDeletedInstance + 1
}

//
// the application wants to know whether this
// peer thinks an instance has been decided,
// and if so what the agreed value is. Status()
// should just inspect the local peer state;
// it should not contact other Paxos peers.
//
func (px *Paxos) Status(seq int) (bool, interface{}) { // TODO make sure only called if Start() called
  px.mu.Lock()
  defer px.mu.Unlock()
  if v, ok := px.instances[seq]; ok {
    return v.Decided, v.DecidedValue
  } else {
    return false, nil
  }
}


//
// tell the peer to shut itself down.
// for testing.
// please do not change this function.
//
func (px *Paxos) Kill() {
  px.dead = true
  if px.l != nil {
    px.l.Close()
  }
}

//
// the application wants to create a paxos peer.
// the ports of all the paxos peers (including this one)
// are in peers[]. this servers port is peers[me].
//
func Make(peers []string, me int, rpcs *rpc.Server) *Paxos {
  px := &Paxos{}
  px.peers = peers
  px.me = me

  px.instances = make(map[int]*InstanceState)
  px.latestDone = make([]int, len(peers))
  for i := 0; i < len(peers); i++ {
    px.latestDone[i] = -1
  }
  px.lastDeletedInstance = -1
  px.highestSeenInstance = -1

  if rpcs != nil {
    // caller will create socket &c
    rpcs.Register(px)
  } else {
    rpcs = rpc.NewServer()
    rpcs.Register(px)

    // prepare to receive connections from clients.
    // change "unix" to "tcp" to use over a network.
    os.Remove(peers[me]) // only needed for "unix"
    l, e := net.Listen("unix", peers[me]);
    if e != nil {
      log.Fatal("listen error: ", e);
    }
    px.l = l
    
    // please do not change any of the following code,
    // or do anything to subvert it.
    
    // create a thread to accept RPC connections
    go func() {
      for px.dead == false {
        conn, err := px.l.Accept()
        if err == nil && px.dead == false {
          if px.unreliable && (rand.Int63() % 1000) < 100 {
            // discard the request.
            conn.Close()
          } else if px.unreliable && (rand.Int63() % 1000) < 200 {
            // process the request but force discard of reply.
            c1 := conn.(*net.UnixConn)
            f, _ := c1.File()
            err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
            if err != nil {
              fmt.Printf("shutdown: %v\n", err)
            }
            px.rpcCount++
            go rpcs.ServeConn(conn)
          } else {
            px.rpcCount++
            go rpcs.ServeConn(conn)
          }
        } else if err == nil {
          conn.Close()
        }
        if err != nil && px.dead == false {
          fmt.Printf("Paxos(%v) accept: %v\n", me, err.Error())
        }
      }
    }()
  }


  return px
}
