package shardkv

import "shardmaster"
import "net/rpc"
import "time"
import "sync"
import "fmt"
import "strconv"

type Clerk struct {
  mu sync.Mutex // one RPC at a time
  sm *shardmaster.Clerk
  config shardmaster.Config
  clientId int64
  requestId int64
  transactionPuts map[int64]map[string]string
}



func MakeClerk(shardmasters []string) *Clerk {
  ck := new(Clerk)
  ck.sm = shardmaster.MakeClerk(shardmasters)
  ck.clientId = nrand()
  ck.requestId = 0
  ck.transactionPuts = make(map[int64]map[string]string)
  return ck
}

//
// call() sends an RPC to the rpcname handler on server srv
// with arguments args, waits for the reply, and leaves the
// reply in reply. the reply argument should be a pointer
// to a reply structure.
//
// the return value is true if the server responded, and false
// if call() was not able to contact the server. in particular,
// the reply's contents are only valid if call() returned true.
//
// you should assume that call() will time out and return an
// error after a while if it doesn't get a reply from the server.
//
// please use call() to send all RPCs, in client.go and server.go.
// please don't change this function.
//
func call(srv string, rpcname string,
          args interface{}, reply interface{}) bool {
  c, errx := rpc.Dial("unix", srv)
  if errx != nil {
    return false
  }
  defer c.Close()
  ch := make(chan error, 1)
  // need timeouts because otherwise deadlock is a possiblity in this design
  go func() { ch <- c.Call(rpcname, args, reply) }()
  select {
    case err := <-ch:
      if err == nil {
        return true
      }
      fmt.Println(err)
      return false
    case <-time.After(1 * time.Second):
      return false
  }
}

//
// which shard is a key in?
// please use this function,
// and please do not change it.
//
func key2shard(key string) int {
  shard := 0
  if len(key) > 0 {
    shard = int(key[0])
  }
  shard %= shardmaster.NShards
  return shard
}

//
// fetch the current value for a key.
// returns "" if the key does not exist.
// keeps trying forever in the face of all other errors.
//
func (ck *Clerk) Get(key string) string {
  ck.mu.Lock()
  defer ck.mu.Unlock()

  ck.requestId += 1

  for {
    shard := key2shard(key)

    gid := ck.config.Shards[shard]

    servers, ok := ck.config.Groups[gid]

    if ok {
      // try each server in the shard's replication group.
      for _, srv := range servers {
        args := &GetArgs{key, ck.clientId, ck.requestId}
        var reply GetReply
        ok := call(srv, "ShardKV.Get", args, &reply)
        if ok && (reply.Err == OK || reply.Err == ErrNoKey) {
          return reply.Value
        }
        if ok && (reply.Err == ErrWrongGroup) {
          break
        }
      }
    }

    time.Sleep(100 * time.Millisecond)

    // ask master for a new configuration.
    ck.config = ck.sm.Query(-1)
  }
  return ""
}

func (ck *Clerk) PutExt(key string, value string, dohash bool) string {
  ck.mu.Lock()
  defer ck.mu.Unlock()

  ck.requestId += 1

  for {
    shard := key2shard(key)

    gid := ck.config.Shards[shard]

    servers, ok := ck.config.Groups[gid]

    if ok {
      // try each server in the shard's replication group.
      for _, srv := range servers {
        args := &PutArgs{key, value, dohash, ck.clientId, ck.requestId}
        var reply PutReply
        ok := call(srv, "ShardKV.Put", args, &reply)
        if ok && reply.Err == OK {
          return reply.PreviousValue
        }
        if ok && (reply.Err == ErrWrongGroup) {
          break
        }
      }
    }

    time.Sleep(100 * time.Millisecond)

    // ask master for a new configuration.
    ck.config = ck.sm.Query(-1)
  }
}

func (ck *Clerk) StartTransaction() Transaction {
  ck.mu.Lock()
  defer ck.mu.Unlock()

  ck.requestId += 1
  transactionID := ck.requestId

  transaction := MakeTransaction(transactionID, ck)
  ck.transactionPuts[transactionID] = make(map[string]string)
  return transaction
}

func (trans *Transaction) PutExt(key string, value string, dohash bool) string {
  prev := ""

  if dohash {
    prev = trans.Get(key)
    value = strconv.Itoa(int(hash(prev + value)))
  }

  trans.ck.mu.Lock()
  trans.ck.transactionPuts[trans.RequestId][key] = value
  shard := key2shard(key)
  trans.ShardOps[shard].Ops = append(trans.ShardOps[shard].Ops, Op{PutType, PutArgs{Key: key, Value: value}})
  trans.ck.mu.Unlock()

  return prev
}

func (trans *Transaction) Put(key string, value string) string {
  return trans.PutExt(key, value, false)
}

func (trans *Transaction) PutHash(key string, value string) string {
  return trans.PutExt(key, value, true)
}

func (trans *Transaction) Get(key string) string {
  transactionID := trans.RequestId
  trans.ck.mu.Lock()
  value, ok := trans.ck.transactionPuts[transactionID][key]
  trans.ck.mu.Unlock()
  if ok {
    return value
  }
  value = trans.ck.Get(key)
  shard := key2shard(key)
  trans.ShardOps[shard].Ops = append(trans.ShardOps[shard].Ops, Op{CheckValueType, CheckValueArgs{key, value}})
  return value
}

func (trans *Transaction) Commit() bool {
  trans.ck.mu.Lock()
  defer trans.ck.mu.Unlock()

  trans.CurrentShard = -1

  shard := nextShard(*trans)
  trans.CurrentShard = shard
  transactionID := trans.RequestId

  for {
    if shard == -1 {
      return true
    }
    // ask master for a new configuration.
    trans.ck.config = trans.ck.sm.Query(-1)

    gid := trans.ck.config.Shards[shard]

    servers, ok := trans.ck.config.Groups[gid]

    if ok {
      // try each server in the shard's replication group.
      for _, srv := range servers {
        args := &ExecuteTransactionArgs{*trans}
        var reply ExecuteTransactionReply
        ok := call(srv, "ShardKV.ExecuteTransaction", args, &reply)
        if ok {
          // Remove the transaction from the map containing transactions
          delete(trans.ck.transactionPuts, transactionID)
          return reply.Success
        }
      }
    }
    time.Sleep(100 * time.Millisecond)

  }
}

func (ck *Clerk) Put(key string, value string) {
  ck.PutExt(key, value, false)
}
func (ck *Clerk) PutHash(key string, value string) string {
  v := ck.PutExt(key, value, true)
  return v
}
