package shardkv
import "hash/fnv"
import "shardmaster"
import "crypto/rand"
import "math/big"

//
// Sharded key/value server.
// Lots of replica groups, each running op-at-a-time paxos.
// Shardmaster decides which group serves each shard.
// Shardmaster may change shard assignment from time to time.
//
// You will have to modify these definitions.
//

const (
  OK = "OK"
  ErrNoKey = "ErrNoKey"
  ErrWrongGroup = "ErrWrongGroup"
)
type Err string

const (
  PutType = "Put"
  GetType = "Get"
  CheckValueType = "CheckValue"
  FPType = "ForwardPass"
  BPType = "BackwardPass"
  TransferType = "Transfer"
  ConfigType = "Config"
  InitiateType = "Initiate"
  EndType = "End"
)

const (
  CommittedState = "Committed"
  AbortedState = "Aborted"
  InFlightState = "InFlight"
)

type OpType string
type InFlightTransactionState string

// Link should contain list of Ops
type Transaction struct {
  RequestId int64
  ClientId int64
  ShardOps [shardmaster.NShards]ShardOps
  Groups [shardmaster.NShards]int64  // the group that handled each shard in the forward pass 
  CurrentShard int
  Dependencies map[RequestKey]DepInfo  // List of TransactionIDs that this transaction has dependencies on
  ck *Clerk  // Clerk that allows transaction objects to process Gets
}

func MakeTransaction(transactionId int64, ck *Clerk) Transaction {
  // ensures that Dependencies is initialized to non-nil value
  return Transaction{Dependencies: make(map[RequestKey]DepInfo),
    RequestId: transactionId, ck: ck, ClientId: ck.clientId}
}

type DepInfo struct {
  ShardOps [shardmaster.NShards]ShardOps
}

type InFlightTransactionInfo struct {
  Type OpType
  Dependencies map[RequestKey]DepInfo
  State InFlightTransactionState
  ShardOps [shardmaster.NShards]ShardOps
}

type RequestKey struct {
  RequestId int64
  ClientId int64
}

type Op struct {
  Type OpType
  Args interface{}
}

type ShardOps struct {
  Ops []Op
  // TODO: Add more ShardOps related things here
}

type PerClientMemory struct {
  MaxRequestId int64
  Result interface{}
}

type PutArgs struct {
  Key string
  Value string
  DoHash bool  // For PutHash
  ClientId int64
  RequestId int64
}

type PutReply struct {
  Err Err
  PreviousValue string   // For PutHash
}

type GetArgs struct {
  Key string
  ClientId int64
  RequestId int64
}

type GetReply struct {
  Err Err
  Value string
}

type CheckValueArgs struct {
  Key string
  Value string
}

type TransferArgs struct {
  ConfigNum int // the config number that the this transfer helps us move towards, should be current
  // config number for the sender and recipient if this transfer is to succeed
  Shard int
  Kv map[string]string
  ClientMax map[int64]PerClientMemory
  ForwardPass map[RequestKey]bool
  BackwardPass map[RequestKey]bool
}

type TransferReply struct {
  Success bool
}

type ForwardPassArgs struct {
  Transaction Transaction
  HasAborted bool
}

type ForwardPassReply struct {
  Success bool // abort if false
}

type BackwardPassArgs struct {
  Transaction Transaction
  ShouldAbort bool
}

type BackwardPassReply struct {}

type ExecuteTransactionArgs struct {
  Transaction Transaction
}

type ExecuteTransactionReply struct {
  Success bool
}

type AbortTransaction struct {
  Transaction Transaction
}

type CommitTransaction struct {
  Transaction Transaction
}

type ConfigProposal struct {
  Conf shardmaster.Config
}

type InitiateTransfer struct {
  ConfigNum int
  Shard int
}

type InitiateTransferReply struct {
  Success bool
}

type EndTransfer struct {
  ConfigNum int
  Shard int
}

func hash(s string) uint32 {
  h := fnv.New32a()
  h.Write([]byte(s))
  return h.Sum32()
}

func nrand() int64 {
  max := big.NewInt(int64(1) << 62)
  bigx, _ := rand.Int(rand.Reader, max)
  x := bigx.Int64()
  return x
}
