package shardkv

import "net"
import "fmt"
import "net/rpc"
import "log"
import "time"
import "paxos"
import "sync"
import "os"
import "syscall"
import "encoding/gob"
import "math/rand"
import "shardmaster"
import "strconv"

const Debug=0

func DPrintf(format string, a ...interface{}) (n int, err error) {
        if Debug > 0 {
                log.Printf(format, a...)
        }
        return
}

type ShardState string
const (
  NotOwned = "NotOwned" // some other group owns it
  Owned = "Owned" // we own it
  Transferring = "Transferring" // we still own it, but it's in the process of being transferred
)

type ShardKV struct {
  mu sync.Mutex
  l net.Listener
  me int
  dead bool // for testing
  unreliable bool // for testing
  sm *shardmaster.Clerk
  px *paxos.Paxos

  gid int64 // my replica group ID

  minUnknown int // minimum seq whose result is unknown
  kv [shardmaster.NShards]map[string]string // kv database
  // TODO move clientMax and kv into a single struct representing the per-shard data
  clientMax [shardmaster.NShards]map[int64]PerClientMemory // max RequestId we've seen from each client,
  // as well as the result for that request
  inFlightTransactions [shardmaster.NShards]map[string]map[RequestKey]InFlightTransactionInfo
  // Keeps track of in-flight transactions on a per-key basis -- maps shardID to a map
  // that maps key to a map of Transaction IDs to some transaction info 
  forwardPass [shardmaster.NShards]map[RequestKey]bool
  backwardPass [shardmaster.NShards]map[RequestKey]bool
  shardStates [shardmaster.NShards]ShardState
  currentConfig shardmaster.Config
}

func (kv *ShardKV) executePut(args PutArgs) PutReply {
  shard := key2shard(args.Key)
  reply := PutReply{}
  if kv.shardStates[shard] != Owned {
    reply.Err = ErrWrongGroup
    return reply // don't save reply in per-client memory
  }

  requestKey := RequestKey{args.RequestId, args.ClientId}
  deps := kv.getSingleKeyDeps(shard, args.Key)
  transactionsForKey, ok := kv.inFlightTransactions[shard][args.Key]
  if !ok {
    transactionsForKey = make(map[RequestKey]InFlightTransactionInfo)
    kv.inFlightTransactions[shard][args.Key] = transactionsForKey
  }
  // don't need to worry about deps here since all deps are for the same key
  // and will be seen anyway
  info := InFlightTransactionInfo{Type: PutType, Dependencies: make(map[RequestKey]DepInfo), State: InFlightState}
  info.ShardOps[shard] = ShardOps{[]Op{Op{PutType, PutArgs{Key: args.Key}}}}
  transactionsForKey[requestKey] = info
  kv.waitForSingleKeyDeps(shard, args.Key, deps)
  info.State = CommittedState
  transactionsForKey[requestKey] = info

  // need to check again since we ceded the lock
  if red, result := kv.isRedundant(args.ClientId, args.RequestId, shard); red {
    if result == nil {
      return PutReply{} // doesn't matter what we return in this case, client has moved on
    } else {
      return result.(PutReply)
    }
  }
  prev := kv.kv[shard][args.Key]
  if args.DoHash {
    args.Value = strconv.Itoa(int(hash(prev + args.Value)))
  }
  kv.kv[shard][args.Key] = args.Value
  reply.Err = OK
  reply.PreviousValue = prev
  kv.clientMax[shard][args.ClientId] = PerClientMemory{args.RequestId, reply}
  return reply
}

func (kv *ShardKV) getSingleKeyDeps(shard int, key string) map[RequestKey]DepInfo {
  deps := make(map[RequestKey]DepInfo)
  // no need for transitive dependencies here since single-key operation
  for reqKey, transInfo := range kv.inFlightTransactions[shard][key] {
    deps[reqKey] = DepInfo{transInfo.ShardOps}
  }
  return deps
}

// lock should be held when calling this method, will be held on return
func (kv *ShardKV) waitForSingleKeyDeps(shard int, key string, deps map[RequestKey]DepInfo) {
  kv.mu.Unlock()
  to := 10 * time.Millisecond
  for !kv.dead {
    kv.mu.Lock()
    if kv.singleKeyDepsComplete(shard, key, deps) {
      return
    }
    kv.mu.Unlock()
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
}

func (kv *ShardKV) executeGet(args GetArgs) GetReply {
  shard := key2shard(args.Key)
  reply := GetReply{}
  if kv.shardStates[shard] != Owned {
    reply.Err = ErrWrongGroup
    return reply // don't save reply in per-client memory
  }

  requestKey := RequestKey{args.RequestId, args.ClientId}
  deps := kv.getSingleKeyDeps(shard, args.Key)
  transactionsForKey, ok := kv.inFlightTransactions[shard][args.Key]
  if !ok {
    transactionsForKey = make(map[RequestKey]InFlightTransactionInfo)
    kv.inFlightTransactions[shard][args.Key] = transactionsForKey
  }
  // don't need to worry about deps here since all deps are for the same key
  // and will be seen anyway
  info := InFlightTransactionInfo{Type: GetType, Dependencies: make(map[RequestKey]DepInfo), State: InFlightState}
  info.ShardOps[shard] = ShardOps{[]Op{Op{GetType, GetArgs{Key: args.Key}}}}
  transactionsForKey[requestKey] = info
  kv.waitForSingleKeyDeps(shard, args.Key, deps)
  info.State = CommittedState
  transactionsForKey[requestKey] = info

  // need to check again since we ceded the lock
  if red, result := kv.isRedundant(args.ClientId, args.RequestId, shard); red {
    if result == nil {
      return GetReply{} // doesn't matter what we return in this case, client has moved on
    } else {
      return result.(GetReply)
    }
  }

  if v, ok := kv.kv[shard][args.Key]; ok {
    reply.Err = OK
    reply.Value = v
  } else {
    reply.Err = ErrNoKey
  }
  kv.clientMax[shard][args.ClientId] = PerClientMemory{args.RequestId, reply}
  return reply
}

func (kv *ShardKV) executeForwardPass(args ForwardPassArgs) ForwardPassReply {
  transaction := args.Transaction
  shard := transaction.CurrentShard
  if kv.shardStates[shard] != Owned {
    return ForwardPassReply{false}
  }
  if _, ok := kv.forwardPass[shard][RequestKey{transaction.RequestId, transaction.ClientId}]; ok {
    return ForwardPassReply{true}
  }
  isValid := !args.HasAborted
  currentShardOps := transaction.ShardOps[shard]
  if isValid {
    // TODO: Optimize this loop a bit by breaking immediately after setting isValid to false
    for _, op := range currentShardOps.Ops {
      if op.Type == CheckValueType {
        args := op.Args.(CheckValueArgs)
        key := args.Key
        if kv.kv[shard][key] != args.Value {
          // Some CheckValue failed -- transaction must be aborted
          isValid = false
        }
        openTransactions, ok := kv.inFlightTransactions[shard][key]
        if ok {
          // If any of the inflight transactions write to this key, this transaction isn't valid
          for _, trans := range openTransactions {
            if trans.Type == PutType && trans.State == InFlightState {
              isValid = false
            }
          }
        }
      }
    }
  }
  transaction.Groups[shard] = kv.gid
  transRequestKey := RequestKey{transaction.RequestId, transaction.ClientId}
  for _, op := range currentShardOps.Ops {
    var key string
    if op.Type == CheckValueType {
      args := op.Args.(CheckValueArgs)
      key = args.Key
    } else if op.Type == PutType {
      args := op.Args.(PutArgs)
      key = args.Key
    }

    transactionsForKey, ok := kv.inFlightTransactions[shard][key]
    if !ok {
      transactionsForKey = make(map[RequestKey]InFlightTransactionInfo)
      kv.inFlightTransactions[shard][key] = transactionsForKey
    }
    if op.Type == PutType && isValid { // don't add dependencies for CheckValueType
      // also don't need to add dependencies if you're going to be aborted
      for reqKey, transInfo := range transactionsForKey {
        // Add dependency information here
        if reqKey != transRequestKey && transInfo.State == InFlightState {
          transaction.Dependencies[reqKey] = DepInfo{transInfo.ShardOps}
          for dep, depInfo := range transInfo.Dependencies {
            if dep != transRequestKey {
              transaction.Dependencies[dep] = depInfo
            }
          }
        }
      }
    }
    transInfo, ok := transactionsForKey[transRequestKey]
    if !ok || transInfo.Type != PutType {
      // Add in-flight transaction information here
      if isValid {
        transactionsForKey[transRequestKey] = InFlightTransactionInfo{op.Type, transaction.Dependencies, InFlightState, transaction.ShardOps}
      } else {
        // if it's going to be aborted, might as well mark it as aborted now so that
        // no one depends on it
        transactionsForKey[transRequestKey] = InFlightTransactionInfo{op.Type, make(map[RequestKey]DepInfo), AbortedState, transaction.ShardOps}
      }
    }
  }
  kv.forwardPass[shard][RequestKey{transaction.RequestId, transaction.ClientId}] = true
  go kv.attemptNextForwardLink(transaction, isValid)
  return ForwardPassReply{true}
}

func nextShard(t Transaction) int {
  currentShard := t.CurrentShard + 1
  for ; currentShard < len(t.ShardOps) && t.ShardOps[currentShard].Ops == nil; currentShard++ {}
  if currentShard == len(t.ShardOps) {
    return -1
  } else {
    return currentShard
  }
}

func previousShard(t Transaction) int {
  currentShard := t.CurrentShard - 1
  for ; currentShard >= 0 && t.ShardOps[currentShard].Ops == nil; currentShard-- {}
  if currentShard < 0 {
    return -1
  } else {
    return currentShard
  }
}

func (kv *ShardKV) attemptNextForwardLink(t Transaction, isValid bool) {
  kv.mu.Lock()
  conf := kv.currentConfig
  kv.mu.Unlock()
  shard := nextShard(t)
  if shard == -1 {
    // Since we successfully reached the last server, we are good
    // and can start the BackwardPass
    reply := &BackwardPassReply{}
    kv.BackwardPass(&BackwardPassArgs{t, !isValid}, reply)
    return
  }
  t.CurrentShard = shard
  reply := &ForwardPassReply{}
  success := false
  // find the replica group that controls the next shard (this assumes that new
  // replica groups cannot join after transactions have started)
  for !success && !kv.dead {
    for _, srvs := range conf.Groups {
      for _, srv := range srvs {
        ok := call(srv, "ShardKV.ForwardPass", &ForwardPassArgs{t, !isValid}, reply)
        if ok && reply.Success {
          success = true
          break
        }
      }
      if success {
        break
      }
    }
  }
}

func (kv *ShardKV) removeInFlightEntries(transaction Transaction, state InFlightTransactionState) {
  transReqKey := RequestKey{transaction.RequestId, transaction.ClientId}
  currentShardOps := transaction.ShardOps[transaction.CurrentShard]
  for _, op := range currentShardOps.Ops {
    var key string
    if op.Type == CheckValueType {
      args := op.Args.(CheckValueArgs)
      key = args.Key
    } else if op.Type == PutType {
      args := op.Args.(PutArgs)
      key = args.Key
    }
    transForKey := kv.inFlightTransactions[transaction.CurrentShard][key]
    entry := transForKey[transReqKey]
    entry.State = state
    transForKey[transReqKey] = entry
  }
}

// called with lock held, should return with lock held
func (kv *ShardKV) abortTransaction(transaction Transaction) {
  shard := transaction.CurrentShard
  kv.removeInFlightEntries(transaction, AbortedState)
  kv.backwardPass[shard][RequestKey{transaction.RequestId, transaction.ClientId}] = true
  go kv.attemptNextBackwardLink(transaction, true)
}

func (kv *ShardKV) shardInFlightTransactions(shard int) bool {
  shardTransactions := kv.inFlightTransactions[shard]
  for _, keyTransactions := range shardTransactions {
    for _, transInfo := range keyTransactions {
      if transInfo.State == InFlightState {
        return true
      }
    }
  }
  return false
}

func (kv *ShardKV) hasOpForKey(so ShardOps, key string) bool {
  if so.Ops == nil {
    return false
  }
  for _, op := range so.Ops {
    if op.Type == GetType && op.Args.(GetArgs).Key == key {
      return true
    } else if op.Type == PutType && op.Args.(PutArgs).Key == key {
      return true
    } else if op.Type == CheckValueType && op.Args.(CheckValueArgs).Key == key {
      return true
    }
  }
  return false
}

func (kv *ShardKV) singleKeyDepsComplete(shard int, key string, dependencies map[RequestKey]DepInfo) bool {
  transactionsForKey := kv.inFlightTransactions[shard][key]
  for transReqKey, depInfo := range dependencies {
    if !kv.hasOpForKey(depInfo.ShardOps[shard], key) {
      continue
    }
    if transInfo, ok := transactionsForKey[transReqKey]; !ok || transInfo.State == InFlightState {
      return false
    }
  }
  return true
}

func (kv *ShardKV) dependenciesComplete(transaction Transaction) bool {
  dependencies := transaction.Dependencies
  shard := transaction.CurrentShard
  currentShardOps := transaction.ShardOps[shard]
  for _, op := range currentShardOps.Ops {
    if op.Type == CheckValueType { // no dependencies for CheckValueType
      continue
    }
    args := op.Args.(PutArgs)
    key := args.Key
    if !kv.singleKeyDepsComplete(shard, key, dependencies) {
      return false
    }
  }
  return true
}

// called with lock held, should be held afterwards
func (kv *ShardKV) commitTransaction(transaction Transaction) {
  shard := transaction.CurrentShard
  currentShardOps := transaction.ShardOps[shard]
  kv.backwardPass[shard][RequestKey{transaction.RequestId, transaction.ClientId}] = true
  to := 10 * time.Millisecond
  kv.mu.Unlock()
  for !kv.dead {
    kv.mu.Lock()
    if kv.dependenciesComplete(transaction) {
      for _, op := range currentShardOps.Ops {
        if op.Type == PutType {
          args := op.Args.(PutArgs)
          // we don't need duplicate detection here since we do that already for the
          // containing operation (the transaction)
          kv.kv[shard][args.Key] = args.Value
        }
      }
      break
    }
    kv.mu.Unlock()
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
  kv.removeInFlightEntries(transaction, CommittedState)
  go kv.attemptNextBackwardLink(transaction, false)
}

func (kv *ShardKV) attemptNextBackwardLink(t Transaction, ShouldAbort bool) {
  kv.mu.Lock()
  conf := kv.currentConfig
  kv.mu.Unlock()
  shard := previousShard(t)
  if shard == -1 {
    return
  }
  t.CurrentShard = shard
  reply := &BackwardPassReply{} // nothing in this struct
  success := false
  for !success && !kv.dead {
    // Need to keep looping on this one since we have to commit or abort
    for _, srv := range conf.Groups[t.Groups[shard]] {
      ok := call(srv, "ShardKV.BackwardPass", &BackwardPassArgs{t, ShouldAbort}, reply)
      if ok {
        success = true
        break
      }
    }
  }
}

func (kv *ShardKV) executeBackwardPass(args BackwardPassArgs) BackwardPassReply {
  transaction := args.Transaction
  shard := transaction.CurrentShard
  if _, ok := kv.backwardPass[shard][RequestKey{transaction.RequestId, transaction.ClientId}]; ok {
    return BackwardPassReply{}
  }
  if args.ShouldAbort {
    kv.abortTransaction(args.Transaction)
  } else {
    kv.commitTransaction(args.Transaction)
  }
  return BackwardPassReply{}
}

// returns whether we've seen a higher requestId from this client, in which
// case the client has already received the desired response for this request.
// the special case is when we have not seen a higher requestId from this client,
// but we have seen the same one -- in this case we also return true as well
// as the result for the request
func (kv *ShardKV) isRedundant(clientId int64, requestId int64, shard int) (bool, interface{}) {
  if clientMax, ok := kv.clientMax[shard][clientId]; !ok || requestId > clientMax.MaxRequestId {
    return false, nil
  } else if requestId == clientMax.MaxRequestId {
    return true, clientMax.Result
  } else {
    return true, nil
  }
}

func (kv *ShardKV) executeTransfer(args TransferArgs) TransferReply {
  // we want reach the state specified by the current config before accepting
  // transfers for higher configs
  if args.ConfigNum > kv.currentConfig.Num {
    return TransferReply{false}
  }
  // transfer must have completed already
  if args.ConfigNum < kv.currentConfig.Num || kv.shardStates[args.Shard] != NotOwned {
    return TransferReply{true}
  }
  kv.kv[args.Shard] = args.Kv
  kv.clientMax[args.Shard] = args.ClientMax
  kv.shardStates[args.Shard] = Owned
  kv.forwardPass[args.Shard] = args.ForwardPass
  kv.backwardPass[args.Shard] = args.BackwardPass
  return TransferReply{true}
}

func (kv *ShardKV) executeConfigChange(args ConfigProposal) {
  if args.Conf.Num != kv.currentConfig.Num + 1 {
    return
  }
  for i := 0; i < shardmaster.NShards; i++ {
    if (kv.currentConfig.Shards[i] == kv.gid && kv.shardStates[i] != Owned) ||
      (kv.currentConfig.Shards[i] != kv.gid && kv.shardStates[i] != NotOwned) {
      return
    }
  }

  kv.currentConfig = args.Conf
  // speical case for first config -- the shard isn't transferred to us;
  // we must create it instead
  if kv.currentConfig.Num == 1 {
    for i := 0; i < shardmaster.NShards; i++ {
      if kv.currentConfig.Shards[i] == kv.gid {
        kv.shardStates[i] = Owned
      }
    }
  }
}

func (kv *ShardKV) executeInitiateTransfer(args InitiateTransfer) InitiateTransferReply {
  if args.ConfigNum != kv.currentConfig.Num || kv.shardStates[args.Shard] == NotOwned ||
  kv.shardInFlightTransactions(args.Shard) {
    return InitiateTransferReply{false}
  }
  kv.shardStates[args.Shard] = Transferring
  return InitiateTransferReply{true}
}

func (kv *ShardKV) executeEndTransfer(args EndTransfer) {
  if args.ConfigNum != kv.currentConfig.Num || kv.shardStates[args.Shard] == NotOwned {
    return
  }
  kv.shardStates[args.Shard] = NotOwned
}

func (kv *ShardKV) applyOp(op Op) interface{} {
  if op.Type == PutType {
    args := op.Args.(PutArgs)
    shard := key2shard(args.Key)
    if red, result := kv.isRedundant(args.ClientId, args.RequestId, shard); red {
      return result
    } else {
      return kv.executePut(args)
    }
  } else if op.Type == GetType {
    args := op.Args.(GetArgs)
    shard := key2shard(args.Key)
    if red, result := kv.isRedundant(args.ClientId, args.RequestId, shard); red {
      return result
    } else {
      return kv.executeGet(args)
    }
  } else if op.Type == FPType {
    args := op.Args.(ForwardPassArgs)
    return kv.executeForwardPass(args)
  } else if op.Type == BPType {
    args := op.Args.(BackwardPassArgs)
    return kv.executeBackwardPass(args)
  } else if op.Type == TransferType {
    args := op.Args.(TransferArgs)
    return kv.executeTransfer(args)
  } else if op.Type == ConfigType {
    args := op.Args.(ConfigProposal)
    kv.executeConfigChange(args)
    return nil
  } else if op.Type == InitiateType {
    args := op.Args.(InitiateTransfer)
    return kv.executeInitiateTransfer(args)
  } else { // EndType
    args := op.Args.(EndTransfer)
    kv.executeEndTransfer(args)
    return nil
  }
}

func opsEqual(op1 Op, op2 Op) bool {
  if op1.Type != op2.Type {
    return false
  }

  if op1.Type == ConfigType {
    a := op1.Args.(ConfigProposal)
    b := op2.Args.(ConfigProposal)
    return a.Conf.Num == b.Conf.Num
  } else if op1.Type == TransferType {
    a := op1.Args.(TransferArgs)
    b := op2.Args.(TransferArgs)
    return a.ConfigNum == b.ConfigNum && a.Shard == b.Shard
  } else if op1.Type == FPType {
    a := op1.Args.(ForwardPassArgs).Transaction
    b := op2.Args.(ForwardPassArgs).Transaction
    return a.RequestId == b.RequestId && a.ClientId == b.ClientId && a.CurrentShard == b.CurrentShard
  } else if op1.Type == BPType {
    a := op1.Args.(BackwardPassArgs).Transaction
    b := op2.Args.(BackwardPassArgs).Transaction
    return a.RequestId == b.RequestId && a.ClientId == b.ClientId && a.CurrentShard == b.CurrentShard
  } else {
    return op1 == op2
  }
}

// returns whether we successfully decided on op at this instance, and the
// result of the operation decided on for this instance (whether or not it
// was the requested operation)
func (kv *ShardKV) tryInstance(seq int, op Op) (bool, interface{}) {
  kv.px.Start(seq, op)

  to := 10 * time.Millisecond
  for !kv.dead {
    decided, v := kv.px.Status(seq)
    if decided {
      result := v.(Op)
      reply := kv.applyOp(result)
      return opsEqual(result, op), reply
    }
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
  return false, nil
}

func (kv *ShardKV) tryTillSuccess(op Op) interface{} {
  for !kv.dead {
    seqNum := kv.minUnknown
    kv.minUnknown += 1
    success, result := kv.tryInstance(seqNum, op)
    kv.px.Done(seqNum)
    if success {
      return result
    }
  }
  return nil
}


func (kv *ShardKV) Get(args *GetArgs, reply *GetReply) error {
  kv.mu.Lock()
  defer kv.mu.Unlock()
  result := kv.tryTillSuccess(Op{GetType, *args})
  if result != nil {
    *reply = result.(GetReply)
  }
  // else this was an old operation for which the client 
  // already has a result and so the client has closed 
  // the connection
  return nil
}

func (kv *ShardKV) Put(args *PutArgs, reply *PutReply) error {
  kv.mu.Lock()
  defer kv.mu.Unlock()
  result := kv.tryTillSuccess(Op{PutType, *args})
  if result != nil {
    *reply = result.(PutReply)
  }
  // else this was an old operation for which the client 
  // already has a result and so the client has closed 
  // the connection
  return nil
}

func (kv *ShardKV) Transfer(args *TransferArgs, reply *TransferReply) error {
  kv.mu.Lock()
  defer kv.mu.Unlock()
  result := kv.tryTillSuccess(Op{TransferType, *args})
  if result != nil {
    *reply = result.(TransferReply)
  }
  return nil
}

func (kv *ShardKV) attemptRemainingTransfers() {
  for i := 0; i < shardmaster.NShards; i++ {
    if kv.currentConfig.Shards[i] != kv.gid && kv.shardStates[i] != NotOwned {
      reply := InitiateTransferReply{}
      result := kv.tryTillSuccess(Op{InitiateType, InitiateTransfer{kv.currentConfig.Num, i}})
      if result != nil {
        reply = result.(InitiateTransferReply)
      }
      if !reply.Success {
        continue
      }
      destServers := kv.currentConfig.Groups[kv.currentConfig.Shards[i]]
      transferReply := &TransferReply{}
      transferArgs := &TransferArgs{kv.currentConfig.Num, i, kv.kv[i], kv.clientMax[i], kv.forwardPass[i], kv.backwardPass[i]}
      for _, srv := range destServers {
        ok := call(srv, "ShardKV.Transfer", transferArgs, transferReply)
        if ok && transferReply.Success {
          kv.tryTillSuccess(Op{EndType, EndTransfer{kv.currentConfig.Num, i}})
          break
        }
      }
    }
  }
}

func (kv *ShardKV) ForwardPass(args *ForwardPassArgs, reply *ForwardPassReply) error {
  kv.mu.Lock()
  defer kv.mu.Unlock()
  result := kv.tryTillSuccess(Op{FPType, *args})
  if result != nil {
    *reply = result.(ForwardPassReply)
  }
  return nil
}

func (kv *ShardKV) BackwardPass(args *BackwardPassArgs, reply *BackwardPassReply) error {
  kv.mu.Lock()
  defer kv.mu.Unlock()
  kv.tryTillSuccess(Op{BPType, *args})
  result := kv.tryTillSuccess(Op{BPType, *args})
  if result != nil {
    *reply = result.(BackwardPassReply)
  }
  return nil
}

func (kv *ShardKV) waitForTransactionCommit(t Transaction) bool {
  ops := t.ShardOps[t.CurrentShard].Ops
  first := ops[0] // must have at least one op at this shard if it's here
  var key string
  if first.Type == PutType {
    key = first.Args.(PutArgs).Key
  } else { // CheckValueType
    key = first.Args.(CheckValueArgs).Key
  }
  to := 10 * time.Millisecond
  for !kv.dead {
    kv.mu.Lock()
    info := kv.inFlightTransactions[t.CurrentShard][key][RequestKey{t.RequestId, t.ClientId}]
    kv.mu.Unlock()
    if info.State == AbortedState {
      return false
    } else if info.State == CommittedState {
      return true
    }
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
  return false
}

func (kv *ShardKV) ExecuteTransaction(args *ExecuteTransactionArgs, reply *ExecuteTransactionReply) error {
  // This is the RPC that will be called by the client, this will in turn call the
  // ForwardPass RPC
  // Spawn the first ForwardPass, and then spin until receive BackwardPass
  t := args.Transaction
  fpReply := &ForwardPassReply{}
  kv.ForwardPass(&ForwardPassArgs{t, false}, fpReply)
  if fpReply.Success == true {
    reply.Success = kv.waitForTransactionCommit(t)
  } else {
    reply.Success = false
  }
  return nil
}

//
// Ask the shardmaster if there's a new configuration;
// if so, re-configure.
//
func (kv *ShardKV) tick() {
  // handle initialization case (config 0) separately
  kv.mu.Lock()
  defer kv.mu.Unlock()
  config := kv.sm.Query(kv.currentConfig.Num + 1)
  kv.tryTillSuccess(Op{ConfigType, ConfigProposal{config}})
  kv.attemptRemainingTransfers()
}


// tell the server to shut itself down.
func (kv *ShardKV) kill() {
  kv.dead = true
  kv.l.Close()
  kv.px.Kill()
}

//
// Start a shardkv server.
// gid is the ID of the server's replica group.
// shardmasters[] contains the ports of the
//   servers that implement the shardmaster.
// servers[] contains the ports of the servers
//   in this replica group.
// Me is the index of this server in servers[].
//
func StartServer(gid int64, shardmasters []string,
                 servers []string, me int) *ShardKV {
  gob.Register(Op{})
  gob.Register(PutArgs{})
  gob.Register(GetArgs{})
  gob.Register(TransferArgs{})
  gob.Register(ConfigProposal{})
  gob.Register(InitiateTransfer{})
  gob.Register(EndTransfer{})
  gob.Register(PerClientMemory{})
  gob.Register(PutReply{})
  gob.Register(GetReply{})
  gob.Register(TransferReply{})
  gob.Register(ForwardPassArgs{})
  gob.Register(ForwardPassReply{})
  gob.Register(BackwardPassArgs{})
  gob.Register(BackwardPassReply{})
  gob.Register(CheckValueArgs{})

  kv := new(ShardKV)
  kv.me = me
  kv.gid = gid
  kv.sm = shardmaster.MakeClerk(shardmasters)

  kv.minUnknown = 0
  for i := 0; i < shardmaster.NShards; i++ {
    kv.kv[i] = make(map[string]string)
    kv.clientMax[i] = make(map[int64]PerClientMemory)
    kv.shardStates[i] = NotOwned
    kv.inFlightTransactions[i] = make(map[string]map[RequestKey]InFlightTransactionInfo)
    kv.forwardPass[i] = make(map[RequestKey]bool)
    kv.backwardPass[i] = make(map[RequestKey]bool)
  }

  rpcs := rpc.NewServer()
  rpcs.Register(kv)

  kv.px = paxos.Make(servers, me, rpcs)


  os.Remove(servers[me])
  l, e := net.Listen("unix", servers[me]);
  if e != nil {
    log.Fatal("listen error: ", e);
  }
  kv.l = l

  // please do not change any of the following code,
  // or do anything to subvert it.

  go func() {
    for kv.dead == false {
      conn, err := kv.l.Accept()
      if err == nil && kv.dead == false {
        if kv.unreliable && (rand.Int63() % 1000) < 100 {
          // discard the request.
          conn.Close()
        } else if kv.unreliable && (rand.Int63() % 1000) < 200 {
          // process the request but force discard of reply.
          c1 := conn.(*net.UnixConn)
          f, _ := c1.File()
          err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
          if err != nil {
            fmt.Printf("shutdown: %v\n", err)
          }
          go rpcs.ServeConn(conn)
        } else {
          go rpcs.ServeConn(conn)
        }
      } else if err == nil {
        conn.Close()
      }
      if err != nil && kv.dead == false {
        fmt.Printf("ShardKV(%v) accept: %v\n", me, err.Error())
        kv.kill()
      }
    }
  }()

  go func() {
    for kv.dead == false {
      kv.tick()
      time.Sleep(250 * time.Millisecond)
    }
  }()

  return kv
}
