package shardkv

import "testing"
import "shardmaster"
import "runtime"
import "strconv"
import "os"
import "time"
import "fmt"
import "sync"
import "math/rand"

func port(tag string, host int) string {
  s := "/var/tmp/824-"
  s += strconv.Itoa(os.Getuid()) + "/"
  os.Mkdir(s, 0777)
  s += "skv-"
  s += strconv.Itoa(os.Getpid()) + "-"
  s += tag + "-"
  s += strconv.Itoa(host)
  return s
}

func NextValue(hprev string, val string) string {
  h := hash(hprev + val)
  return strconv.Itoa(int(h))
}

func mcleanup(sma []*shardmaster.ShardMaster) {
  for i := 0; i < len(sma); i++ {
    if sma[i] != nil {
      sma[i].Kill()
    }
  }
}

func cleanup(sa [][]*ShardKV) {
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      sa[i][j].kill()
    }
  }
}

func setup(tag string, unreliable bool) ([]string, []int64, [][]string, [][]*ShardKV, func()) {
  runtime.GOMAXPROCS(4)
  
  const nmasters = 3
  var sma []*shardmaster.ShardMaster = make([]*shardmaster.ShardMaster, nmasters)
  var smh []string = make([]string, nmasters)
  // defer mcleanup(sma)
  for i := 0; i < nmasters; i++ {
    smh[i] = port(tag+"m", i)
  }
  for i := 0; i < nmasters; i++ {
    sma[i] = shardmaster.StartServer(smh, i)
  }

  const ngroups = 3   // replica groups
  const nreplicas = 3 // servers per group
  gids := make([]int64, ngroups)    // each group ID
  ha := make([][]string, ngroups)   // ShardKV ports, [group][replica]
  sa := make([][]*ShardKV, ngroups) // ShardKVs
  // defer cleanup(sa)
  for i := 0; i < ngroups; i++ {
    gids[i] = int64(i + 100)
    sa[i] = make([]*ShardKV, nreplicas)
    ha[i] = make([]string, nreplicas)
    for j := 0; j < nreplicas; j++ {
      ha[i][j] = port(tag+"s", (i*nreplicas)+j)
    }
    for j := 0; j < nreplicas; j++ {
      sa[i][j] = StartServer(gids[i], smh, ha[i], j)
      sa[i][j].unreliable = unreliable
    }
  }

  clean := func() { cleanup(sa) ; mcleanup(sma) }
  return smh, gids, ha, sa, clean
}

func TestBasic(t *testing.T) {
  smh, gids, ha, _, clean := setup("basic", false)
  defer clean()

  fmt.Printf("Test: Basic Join/Leave ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(smh)

  ck.Put("a", "x")
  v := ck.PutHash("a", "b")
  if v != "x" {
    t.Fatalf("Puthash got wrong value")
  }
  ov := NextValue("x", "b")
  if ck.Get("a") != ov {
    t.Fatalf("Get got wrong value")
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  // are keys still there after joins?
  for g := 1; g < len(gids); g++ {
    mck.Join(gids[g], ha[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("joining; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }
  
  // are keys still there after leaves?
  for g := 0; g < len(gids)-1; g++ {
    mck.Leave(gids[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("leaving; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  fmt.Printf("  ... Passed\n")
}

func TestMove(t *testing.T) {
  smh, gids, ha, _, clean := setup("move", false)
  defer clean()

  fmt.Printf("Test: Shards really move ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(smh)

  // insert one key per shard
  for i := 0; i < shardmaster.NShards; i++ {
    ck.Put(string('0'+i), string('0'+i))
  }

  // add group 1.
  mck.Join(gids[1], ha[1])
  time.Sleep(5 * time.Second)
  
  // check that keys are still there.
  for i := 0; i < shardmaster.NShards; i++ {
    if ck.Get(string('0'+i)) != string('0'+i) {
      t.Fatalf("missing key/value")
    }
  }

  // remove sockets from group 0.
  for i := 0; i < len(ha[0]); i++ {
    os.Remove(ha[0][i])
  }

  count := 0
  var mu sync.Mutex
  for i := 0; i < shardmaster.NShards; i++ {
    go func(me int) {
      myck := MakeClerk(smh)
      v := myck.Get(string('0'+me))
      if v == string('0'+me) {
        mu.Lock()
        count++
        mu.Unlock()
      } else {
        t.Fatalf("Get(%v) yielded %v\n", i, v)
      }
    }(i)
  }

  time.Sleep(10 * time.Second)

  if count > shardmaster.NShards / 3 && count < 2*(shardmaster.NShards/3) {
    fmt.Printf("  ... Passed\n")
  } else {
    t.Fatalf("%v keys worked after killing 1/2 of groups; wanted %v",
      count, shardmaster.NShards / 2)
  }
}

func TestLimp(t *testing.T) {
  smh, gids, ha, sa, clean := setup("limp", false)
  defer clean()

  fmt.Printf("Test: Reconfiguration with some dead replicas ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(smh)

  ck.Put("a", "b")
  if ck.Get("a") != "b" {
    t.Fatalf("got wrong value")
  }

  for g := 0; g < len(sa); g++ {
    sa[g][rand.Int() % len(sa[g])].kill()
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  // are keys still there after joins?
  for g := 1; g < len(gids); g++ {
    mck.Join(gids[g], ha[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("joining; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }
  
  // are keys still there after leaves?
  for g := 0; g < len(gids)-1; g++ {
    mck.Leave(gids[g])
    time.Sleep(2 * time.Second)
    for i := 0; i < len(sa[g]); i++ {
      sa[g][i].kill()
    }
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("leaving; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  fmt.Printf("  ... Passed\n")
}

func TestTransactionBasic(t *testing.T) {
  smh, gids, ha, _, clean := setup("transactionBasic", false)
  defer clean()

  fmt.Printf("Test: Basic transaction functionality ...\n")

  mck := shardmaster.MakeClerk(smh)
  for i:=0; i < len(gids); i+=1 {
    mck.Join(gids[i], ha[i])
  }

  // wait for config to catch up with joins
  done := false
  for !done {
    config := mck.Query(-1)
    if len(config.Groups) == len(gids) {
      done = true
    }
  }
  time.Sleep(2 * time.Second)

  ck := MakeClerk(smh)
  var transaction Transaction

  // loop until transaction successfully commits
  result := false
  for !result {
    transaction = ck.StartTransaction()
    transactionFailed := false

    key := strconv.Itoa(rand.Int())
    val := strconv.Itoa(rand.Int())
    transaction.Put(key, val)
    newVal := strconv.Itoa(rand.Int())
    v := transaction.PutHash(key, newVal)
    if val != v {
      transactionFailed = true
    }
    for iters:=0; iters < 5; iters+=1 {
      ov := NextValue(val, newVal)
      newVal = strconv.Itoa(rand.Int())
      val = transaction.PutHash(key, newVal)
      if ov != val {
        transactionFailed = true
      }
    }

    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed PutHash test")
    }
  }

  result = false
  for !result {
    transactionFailed := false  // Keeps track of failures within transaction
    // Need to fail the entire transaction if transactionFailed is true and the
    // transaction successfully commits
    transaction = ck.StartTransaction()
    for i := 0; i < shardmaster.NShards; i++ {
      transaction.Put(string('0'+i), string('0'+i))
    }
  
    // check that keys are still there.
    for i := 0; i < shardmaster.NShards; i++ {
      if transaction.Get(string('0'+i)) != string('0'+i) {
        transactionFailed = true
      }
    }

    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed")
    }
  }

  for i := 0; i < shardmaster.NShards; i++ {
    if ck.Get(string('0'+i)) != string('0'+i) {
      t.Fatalf("can't see results of Puts in k/v server")
    }
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  result = false

  for !result {
    transactionFailed := false
    transaction = ck.StartTransaction()
    for i := 0; i < len(keys); i++ {
      keys[i] = strconv.Itoa(rand.Int())
      vals[i] = strconv.Itoa(rand.Int())
      transaction.Put(keys[i], vals[i])
    }

    // do transactions still work with moves?
    for g := 1; g < len(gids); g++ {
      mck.Move(rand.Int() % shardmaster.NShards,
        gids[rand.Int() % len(gids)])
      time.Sleep(1 * time.Second)
      for i := 0; i < len(keys); i++ {
        v := transaction.Get(keys[i])
        if v != vals[i] {
          transactionFailed = true
        }
        vals[i] = strconv.Itoa(rand.Int())
        transaction.Put(keys[i], vals[i])
      }
    }
    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed")
    }
  }

  // make sure transaction puts are successfully recorded
  for i := 0; i < len(keys); i++ {
    v := ck.Get(keys[i])
    if v != vals[i] {
      t.Fatalf("wrong value after transaction; k=%v wanted=%v got=%v",
        keys[i], vals[i], v)
    }
  }

  fmt.Printf("  ... Passed\n")
}

func TestTransactionLimp(t *testing.T) {
  smh, gids, ha, sa, clean := setup("transactionBasic", false)
  defer clean()

  fmt.Printf("Test: Transaction test with some killed replicas ...\n")

  mck := shardmaster.MakeClerk(smh)
  for i:=0; i < len(gids); i+=1 {
    mck.Join(gids[i], ha[i])
  }

  // wait for config to catch up with joins
  done := false
  for !done {
    config := mck.Query(-1)
    if len(config.Groups) == len(gids) {
      done = true
    }
  }
  time.Sleep(2 * time.Second)

  // kill some servers
  for g := 0; g < len(sa); g++ {
    sa[g][rand.Int() % len(sa[g])].kill()
  }

  ck := MakeClerk(smh)
  var transaction Transaction

  // test with putHashes in the transaction
  result := false
  for !result {
    transaction = ck.StartTransaction()
    transactionFailed := false

    key := strconv.Itoa(rand.Int())
    val := strconv.Itoa(rand.Int())
    transaction.Put(key, val)
    newVal := strconv.Itoa(rand.Int())
    v := transaction.PutHash(key, newVal)
    if val != v {
      transactionFailed = true
    }
    for iters:=0; iters < 5; iters+=1 {
      ov := NextValue(val, newVal)
      newVal = strconv.Itoa(rand.Int())
      val = transaction.PutHash(key, newVal)
      if ov != val {
        transactionFailed = true
      }
    }

    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed PutHash test")
    }
  }

  // test with just gets and puts in the transaction
  result = false
  for !result {
    transactionFailed := false  // Keeps track of failures within transaction
    // Need to fail the entire transaction if transactionFailed is true and the
    // transaction successfully commits
    transaction = ck.StartTransaction()
    for i := 0; i < shardmaster.NShards; i++ {
      transaction.Put(string('0'+i), string('0'+i))
    }
  
    // check that keys are still there.
    for i := 0; i < shardmaster.NShards; i++ {
      if transaction.Get(string('0'+i)) != string('0'+i) {
        transactionFailed = true
      }
    }

    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed")
    }
  }

  for i := 0; i < shardmaster.NShards; i++ {
    if ck.Get(string('0'+i)) != string('0'+i) {
      t.Fatalf("can't see results of Puts in k/v server")
    }
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  result = false

  for !result {
    transactionFailed := false
    transaction = ck.StartTransaction()
    for i := 0; i < len(keys); i++ {
      keys[i] = strconv.Itoa(rand.Int())
      vals[i] = strconv.Itoa(rand.Int())
      transaction.Put(keys[i], vals[i])
    }

    // do transactions still work with moves?
    for g := 1; g < len(gids); g++ {
      mck.Move(rand.Int() % shardmaster.NShards,
        gids[rand.Int() % len(gids)])
      time.Sleep(1 * time.Second)
      for i := 0; i < len(keys); i++ {
        v := transaction.Get(keys[i])
        if v != vals[i] {
          transactionFailed = true
        }
        vals[i] = strconv.Itoa(rand.Int())
        transaction.Put(keys[i], vals[i])
      }
    }
    result = transaction.Commit()
    if transactionFailed == true && result == true {
      t.Fatalf("transaction failed")
    }
  }

  // make sure transaction puts are successfully recorded
  for i := 0; i < len(keys); i++ {
    v := ck.Get(keys[i])
    if v != vals[i] {
      t.Fatalf("wrong value after transaction; k=%v wanted=%v got=%v",
        keys[i], vals[i], v)
    }
  }

  fmt.Printf("  ... Passed\n")
}

func getInteger(i string) int {
  integer, e := strconv.Atoi(i)
  if e == nil {
    return integer
  }
  return 0
}

func TestTransactionTestAndSet(t *testing.T) {
  smh, gids, ha, _, clean := setup("transTestAndSet", false)
  defer clean()

  fmt.Printf("Test: Transaction with test-and-sets ...\n")

  mck := shardmaster.MakeClerk(smh)
  for i:=0; i < len(gids); i+=1 {
    mck.Join(gids[i], ha[i])
  }

  done := false
  for !done {
    config := mck.Query(-1)
    if len(config.Groups) == len(gids) {
      done = true
    }
  }
  time.Sleep(2 * time.Second)

  ck := MakeClerk(smh)
  ck.Put("a", "0")

  const npara = 30
  var ca [npara]chan bool
  for i := 0; i < npara; i++ {
    ca[i] = make(chan bool)
    // spawn multiple go routines to simulate Concurrent load
    go func(me int) {
      ok := true
      defer func() { ca[me] <- ok }()
      result := false
      // loop until transaction successfully commits
      for !result {
        transaction := ck.StartTransaction()

        // each thread tests-and-sets the value corresponding to key "a"
        value := getInteger(transaction.Get("a"))
        transaction.Put("a", strconv.Itoa(value + 1))
        result = transaction.Commit()
      }

      time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
    }(i)
  }

  for i := 0; i < npara; i++ {
    x := <- ca[i]
    if x == false {
      t.Fatalf("something is wrong")
    }
  }

  value := ck.Get("a")
  if value != strconv.Itoa(npara) {
    t.Fatalf("transaction test-and-set not working")
  }

  fmt.Printf("  ... Passed\n")
}

func TestTransactionShardInvariants(t *testing.T) {
  smh, gids, ha, _, clean := setup("transInvariants", false)
  defer clean()

  fmt.Printf("Test: Transactions which test invariants across shards ...\n")

  mck := shardmaster.MakeClerk(smh)
  for i:=0; i < len(gids); i+=1 {
    mck.Join(gids[i], ha[i])
  }

  done := false
  for !done {
    config := mck.Query(-1)
    if len(config.Groups) == len(gids) {
      done = true
    }
  }
  time.Sleep(2 * time.Second)

  ck := MakeClerk(smh)

  // Setup invariants here
  // 1. A + B = 10
  // 2. B + C = 32
  // 3. A + C + D = -26
  // 4. D + E = 14
  // 5. E + F = 22
  aKey := strconv.Itoa(rand.Int())
  bKey := strconv.Itoa(rand.Int())
  cKey := strconv.Itoa(rand.Int())
  dKey := strconv.Itoa(rand.Int())
  eKey := strconv.Itoa(rand.Int())
  fKey := strconv.Itoa(rand.Int())

  a := 10
  b := 22
  c := 36
  d := -28
  e := -402
  f := 365

  totalSum := a + b + c + d + e + f

  ck.Put(aKey, strconv.Itoa(a))
  ck.Put(bKey, strconv.Itoa(b))
  ck.Put(cKey, strconv.Itoa(c))
  ck.Put(dKey, strconv.Itoa(d))
  ck.Put(eKey, strconv.Itoa(e))
  ck.Put(fKey, strconv.Itoa(f))

  const npara = 24
  var ca [npara]chan bool
  for i := 0; i < npara; i++ {
    ca[i] = make(chan bool)
    // spawn multiple go routines to simulate Concurrent load
    go func(me int) {
      ok := true
      defer func() { ca[me] <- ok }()
      result := false
      // loop until transaction successfully commits
      for !result {
        transaction := ck.StartTransaction()
        // Perform operations that maintain invariants here

        switch me % 6 {
        case 0:
          a := getInteger(transaction.Get(aKey))
          b := getInteger(transaction.Get(bKey))
          diff := rand.Int()
          transaction.Put(aKey, strconv.Itoa(a + diff))
          transaction.Put(bKey, strconv.Itoa(b - diff))
        case 1:
          b := getInteger(transaction.Get(bKey))
          c := getInteger(transaction.Get(cKey))
          diff := rand.Int()
          transaction.Put(bKey, strconv.Itoa(b - diff))
          transaction.Put(cKey, strconv.Itoa(c + diff))
        case 2:
          a := getInteger(transaction.Get(aKey))
          c := getInteger(transaction.Get(cKey))
          d := getInteger(transaction.Get(dKey))
          newA := rand.Int()
          newC := rand.Int()
          newD := (a + c + d) - (newA + newC)
          transaction.Put(aKey, strconv.Itoa(newA))
          transaction.Put(cKey, strconv.Itoa(newC))
          transaction.Put(dKey, strconv.Itoa(newD))
        case 3:
          d := getInteger(transaction.Get(dKey))
          e := getInteger(transaction.Get(eKey))
          diff := rand.Int()
          transaction.Put(dKey, strconv.Itoa(d + diff))
          transaction.Put(eKey, strconv.Itoa(e - diff))
        case 4:
          e := getInteger(transaction.Get(eKey))
          f := getInteger(transaction.Get(fKey))
          diff := rand.Int()
          transaction.Put(eKey, strconv.Itoa(e - diff))
          transaction.Put(fKey, strconv.Itoa(f + diff))
        case 5:
          a := getInteger(transaction.Get(aKey))
          c := getInteger(transaction.Get(cKey))
          d := getInteger(transaction.Get(dKey))
          f := getInteger(transaction.Get(fKey))
          newA := rand.Int()
          newC := rand.Int()
          newD := 100 - newA
          newF := (a + c + d + f) - (newA + newC + newD)
          transaction.Put(aKey, strconv.Itoa(newA))
          transaction.Put(cKey, strconv.Itoa(newC))
          transaction.Put(dKey, strconv.Itoa(newD))
          transaction.Put(fKey, strconv.Itoa(newF))
        }

        result = transaction.Commit()
      }

      time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
    }(i)
  }

  for i := 0; i < npara; i++ {
    x := <- ca[i]
    if x == false {
      t.Fatalf("something is wrong")
    }
  }

  // Check invariants here
  a = getInteger(ck.Get(aKey))
  b = getInteger(ck.Get(bKey))
  c = getInteger(ck.Get(cKey))
  d = getInteger(ck.Get(dKey))
  e = getInteger(ck.Get(eKey))
  f = getInteger(ck.Get(fKey))

  if a + b + c + d + e + f != totalSum {
    t.Fatalf("invariant broken")
  }

  fmt.Printf("  ... Passed\n")
}

func doConcurrentTransaction(t *testing.T, reliable bool) {
  smh, gids, ha, _, clean := setup("transconc" + strconv.FormatBool(reliable), reliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  for i:=0; i < len(gids); i+=1 {
    mck.Join(gids[i], ha[i])
  }

  done := false
  for !done {
    config := mck.Query(-1)
    if len(config.Groups) == len(gids) {
      done = true
    }
  }
  time.Sleep(2 * time.Second)

  ck := MakeClerk(smh)

  const npara = 11
  var ca [npara]chan bool
  for i := 0; i < npara; i++ {
    ca[i] = make(chan bool)
    // spawn multiple go routines to simulate Concurrent load
    go func(me int) {
      ok := true
      defer func() { ca[me] <- ok }()
      result := false
      // loop until transaction successfully commits
      for !result {
        transaction := ck.StartTransaction()
        // transaction gets and puts
        transactionFailed := false
        mymck := shardmaster.MakeClerk(smh)

        // make sure different threads write to different keys so that
        // results are deterministic
       	key := strconv.Itoa(me)
        last := ""
        for iters := 0; iters < 3; iters++ {
          nv := strconv.Itoa(rand.Int())
          v := transaction.PutHash(key, nv)
          if v != last {
            transactionFailed = true
          }
          last = NextValue(last, nv)
          v = transaction.Get(key)
          if v != last {
            transactionFailed = true
          }

          mymck.Move(rand.Int() % shardmaster.NShards,
            gids[rand.Int() % len(gids)])

          time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
        }
        result = transaction.Commit()
        // if transaction commits but we noticed a failure, need to return failure
        if result == true && transactionFailed == true {
          ok = false
          t.Fatalf("transaction failed")
        }
      }

      time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
    }(i)
  }

  for i := 0; i < npara; i++ {
    x := <- ca[i]
    if x == false {
      t.Fatalf("something is wrong")
    }
  }
}

func TestTransactionConcurrent(t *testing.T) {
  fmt.Printf("Test: Concurrent Transaction ...\n")
  doConcurrentTransaction(t, false)
  fmt.Printf("  ... Passed\n")
}

func TestTransactionConcurrentUnreliable(t *testing.T) {
  fmt.Printf("Test: Concurrent Transaction (unreliable) ...\n")
  doConcurrentTransaction(t, true)
  fmt.Printf("  ... Passed\n")
}

func doConcurrent(t *testing.T, unreliable bool) {
  smh, gids, ha, _, clean := setup("conc"+strconv.FormatBool(unreliable), unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  for i := 0; i < len(gids); i++ {
    mck.Join(gids[i], ha[i])
  }

  const npara = 11
  var ca [npara]chan bool
  for i := 0; i < npara; i++ {
    ca[i] = make(chan bool)
    go func(me int) {
      ok := true
      defer func() { ca[me] <- ok }()
      ck := MakeClerk(smh)
      mymck := shardmaster.MakeClerk(smh)
      key := strconv.Itoa(me)
      last := ""
      for iters := 0; iters < 3; iters++ {
        nv := strconv.Itoa(rand.Int())
        v := ck.PutHash(key, nv)
        if v != last {
          ok = false
          t.Fatalf("PutHash(%v) expected %v got %v\n", key, last, v)
        }
        last = NextValue(last, nv)
        v = ck.Get(key)
        if v != last {
          ok = false
          t.Fatalf("Get(%v) expected %v got %v\n", key, last, v)
        }

        mymck.Move(rand.Int() % shardmaster.NShards,
          gids[rand.Int() % len(gids)])

        time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
      }
    }(i)
  }

  for i := 0; i < npara; i++ {
    x := <- ca[i]
    if x == false {
      t.Fatalf("something is wrong")
    }
  }
}

func TestConcurrent(t *testing.T) {
  fmt.Printf("Test: Concurrent Put/Get/Move ...\n")
  doConcurrent(t, false)
  fmt.Printf("  ... Passed\n")
}

func TestConcurrentUnreliable(t *testing.T) {
  fmt.Printf("Test: Concurrent Put/Get/Move (unreliable) ...\n")
  doConcurrent(t, true)
  fmt.Printf("  ... Passed\n")
}
