package shardmaster

import "net"
import "fmt"
import "net/rpc"
import "log"
import "paxos"
import "sync"
import "os"
import "syscall"
import "encoding/gob"
import "math/rand"
import "sort"
import "time"

type ShardMaster struct {
  mu sync.Mutex
  l net.Listener
  me int
  dead bool // for testing
  unreliable bool // for testing
  px *paxos.Paxos

  minUnknown int // minimum seq whose result is unknown
  configs []Config // indexed by config num
}

const (
  JoinType = "Join"
  LeaveType = "Leave"
  MoveType = "Move"
  QueryType = "Query"
)

type OpType string

type Op struct {
  Type OpType
  Args interface{}
}

func (sm *ShardMaster) nextConfig() Config {
  newConfig := sm.configs[len(sm.configs) - 1]
  newConfig.Num++
  oldGroups := newConfig.Groups
  newConfig.Groups = map[int64][]string{}
  for k, v := range oldGroups {
    newConfig.Groups[k] = v
  }
  return newConfig
}

type Group struct {
  id int64
  numShards int
}

type ByShardCount struct {
  groups []Group
  reverse bool
}

func (a ByShardCount) Len() int { return len(a.groups) }
func (a ByShardCount) Swap(i, j int) { a.groups[i], a.groups[j] = a.groups[j], a.groups[i] }
func (a ByShardCount) Less(i, j int) bool {
  if a.reverse {
    if a.groups[i].numShards > a.groups[j].numShards {
      return true
    } else if a.groups[i].numShards == a.groups[j].numShards {
      return a.groups[i].id > a.groups[j].id
    } else {
      return false
    }
  } else {
    if a.groups[i].numShards < a.groups[j].numShards {
      return true
    } else if a.groups[i].numShards == a.groups[j].numShards {
      return a.groups[i].id < a.groups[j].id
    } else {
      return false
    }
  }
}

// assumes that the config has at least one group
func getSortedGroupSlice(conf Config, reverse bool) []Group {
  groupsMap := map[int64]*Group{}
  for gid := range conf.Groups {
    groupsMap[gid] = &Group{gid, 0}
  }
  for _, gid := range conf.Shards {
    if _, ok := groupsMap[gid]; ok {
      groupsMap[gid].numShards++
    }
  }
  groupsSlice := make([]Group, len(groupsMap))
  i := 0
  for _, group := range groupsMap {
    groupsSlice[i] = *group
    i++
  }
  sort.Sort(ByShardCount{groupsSlice, reverse})
  return groupsSlice
}

func (sm *ShardMaster) executeJoin(args JoinArgs) JoinReply {
  conf := sm.nextConfig()
  if len(conf.Groups) == 0 {
    for i := 0; i < NShards; i++ {
      conf.Shards[i] = args.GID
    }
  } else if _, ok := conf.Groups[args.GID]; !ok {
    shardsToTransfer := NShards / (len(conf.Groups) + 1)
    groupSlice := getSortedGroupSlice(conf, true)
    for ; shardsToTransfer > 0; {
      for _, group := range groupSlice {
        if shardsToTransfer == 0 {
          break
        }
        for i := 0; i < NShards; i++ {
          if conf.Shards[i] == group.id {
            conf.Shards[i] = args.GID
            shardsToTransfer--
            break
          }
        }
      }
    }
  }
  conf.Groups[args.GID] = args.Servers
  sm.configs = append(sm.configs, conf)
  return JoinReply{}
}

func (sm *ShardMaster) executeLeave(args LeaveArgs) LeaveReply {
  conf := sm.nextConfig()
  if _, ok := conf.Groups[args.GID]; ok {
    if len(conf.Groups) == 1 {
      defaultConf := Config{}
      defaultConf.Num = conf.Num
      defaultConf.Groups = map[int64][]string{}
      conf = defaultConf
    } else {
      delete(conf.Groups, args.GID)
      groupSlice := getSortedGroupSlice(conf, false)
      sliceInd := 0
      for i := 0; i < NShards; i++ {
        if conf.Shards[i] == args.GID {
          conf.Shards[i] = groupSlice[sliceInd].id
          sliceInd = (sliceInd + 1) % len(groupSlice)
        }
      }
    }
  }
  sm.configs = append(sm.configs, conf)
  return LeaveReply{}
}

func (sm *ShardMaster) executeMove(args MoveArgs) MoveReply {
  conf := sm.nextConfig()
  conf.Shards[args.Shard] = args.GID
  sm.configs = append(sm.configs, conf)
  return MoveReply{}
}

func (sm *ShardMaster) executeQuery(args QueryArgs) QueryReply {
  num := args.Num
  if num == -1 || num >= len(sm.configs) {
    num = len(sm.configs) - 1
  }
  return QueryReply{sm.configs[num]}
}

func (sm *ShardMaster) applyOp(op Op) interface{} {
  if op.Type == JoinType {
    args := op.Args.(JoinArgs)
    return sm.executeJoin(args)
  } else if op.Type == LeaveType {
    args := op.Args.(LeaveArgs)
    return sm.executeLeave(args)
  } else if op.Type == MoveType {
    args := op.Args.(MoveArgs)
    return sm.executeMove(args)
  } else { // QueryType 
    args := op.Args.(QueryArgs)
    return sm.executeQuery(args)
  }
}

func opsEqual(op1 Op, op2 Op) bool {
  if op1.Type != op2.Type {
    return false
  }

  if op1.Type == JoinType {
    a := op1.Args.(JoinArgs)
    b := op2.Args.(JoinArgs)
    return a.GID == b.GID
  } else {
    return op1 == op2
  }
}

// returns whether we successfully decided on op at this instance, and the
// result of the operation decided on for this instance (whether or not it
// was the requested operation)
func (sm *ShardMaster) tryInstance(seq int, op Op) (bool, interface{}) {
  sm.px.Start(seq, op)

  to := 10 * time.Millisecond
  for {
    decided, v := sm.px.Status(seq)
    if decided {
      result := v.(Op)
      reply := sm.applyOp(result)
      return opsEqual(result, op), reply
    }
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
}

func (sm *ShardMaster) tryTillSuccess(op Op) interface{} {
  for {
    success, result := sm.tryInstance(sm.minUnknown, op)
    sm.px.Done(sm.minUnknown)
    sm.minUnknown += 1
    if success {
      return result
    }
  }
}

func (sm *ShardMaster) Join(args *JoinArgs, reply *JoinReply) error {
  sm.mu.Lock()
  defer sm.mu.Unlock()
  result := sm.tryTillSuccess(Op{JoinType, *args})
  *reply = result.(JoinReply)
  return nil
}

func (sm *ShardMaster) Leave(args *LeaveArgs, reply *LeaveReply) error {
  sm.mu.Lock()
  defer sm.mu.Unlock()
  result := sm.tryTillSuccess(Op{LeaveType, *args})
  *reply = result.(LeaveReply)
  return nil
}

func (sm *ShardMaster) Move(args *MoveArgs, reply *MoveReply) error {
  sm.mu.Lock()
  defer sm.mu.Unlock()
  result := sm.tryTillSuccess(Op{MoveType, *args})
  *reply = result.(MoveReply)
  return nil
}

func (sm *ShardMaster) Query(args *QueryArgs, reply *QueryReply) error {
  sm.mu.Lock()
  defer sm.mu.Unlock()
  result := sm.tryTillSuccess(Op{QueryType, *args})
  *reply = result.(QueryReply)
  return nil
}

// please don't change this function.
func (sm *ShardMaster) Kill() {
  sm.dead = true
  sm.l.Close()
  sm.px.Kill()
}

//
// servers[] contains the ports of the set of
// servers that will cooperate via Paxos to
// form the fault-tolerant shardmaster service.
// me is the index of the current server in servers[].
// 
func StartServer(servers []string, me int) *ShardMaster {
  gob.Register(Op{})
  gob.Register(QueryArgs{})
  gob.Register(JoinArgs{})
  gob.Register(LeaveArgs{})
  gob.Register(MoveArgs{})

  sm := new(ShardMaster)
  sm.me = me

  sm.configs = make([]Config, 1)
  sm.configs[0].Groups = map[int64][]string{}

  rpcs := rpc.NewServer()
  rpcs.Register(sm)

  sm.px = paxos.Make(servers, me, rpcs)
  sm.minUnknown = 0

  os.Remove(servers[me])
  l, e := net.Listen("unix", servers[me]);
  if e != nil {
    log.Fatal("listen error: ", e);
  }
  sm.l = l

  // please do not change any of the following code,
  // or do anything to subvert it.

  go func() {
    for sm.dead == false {
      conn, err := sm.l.Accept()
      if err == nil && sm.dead == false {
        if sm.unreliable && (rand.Int63() % 1000) < 100 {
          // discard the request.
          conn.Close()
        } else if sm.unreliable && (rand.Int63() % 1000) < 200 {
          // process the request but force discard of reply.
          c1 := conn.(*net.UnixConn)
          f, _ := c1.File()
          err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
          if err != nil {
            fmt.Printf("shutdown: %v\n", err)
          }
          go rpcs.ServeConn(conn)
        } else {
          go rpcs.ServeConn(conn)
        }
      } else if err == nil {
        conn.Close()
      }
      if err != nil && sm.dead == false {
        fmt.Printf("ShardMaster(%v) accept: %v\n", me, err.Error())
        sm.Kill()
      }
    }
  }()

  return sm
}
